<?php
/*
 * ZnetDK, Starter Web Application for rapid & easy development
 * See official website http://www.znetdk.fr 
 * Copyright (C) 2015 Pascal MARTINEZ (contact@znetdk.fr)
 * License GNU GPL http://www.gnu.org/licenses/gpl-3.0.html GNU GPL
 * --------------------------------------------------------------------
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * --------------------------------------------------------------------
 * Core renderer of the ressource dependencies  
 *
 * File version: 1.8
 * Last update: 09/03/2021
 */

/**
 * Adds the HTML dependencies to the page layout that are required by ZnetDK
 */
class Dependencies {

    /**
     * Main public method called by the Layout controller to insert the 
     * HTML dependencies to the main page of the application.
     */
    public static function render() {
        $layout = \Parameters::getPageLayoutName();
        if ($layout === 'mobile') {
            // CSS rendering
            self::renderCSSforMobile();
            // JS rendering
            self::renderJSforMobile();
        } else {
            if (CFG_NON_MOBILE_PWA_ENABLED === FALSE) {
                $icon = defined('LC_HEAD_ICO_LOGO') ? LC_HEAD_ICO_LOGO : LC_HEAD_IMG_LOGO;
                echo '<link rel="icon" type="image/png" href="'.$icon.'">'.PHP_EOL;
            }
            self::renderCSS();
            self::renderJS();
        }
    }
    
    /**
     * Renders the CSS dependencies
     */
    static private function renderCSS() {
        $activeTheme = ThemeManager::getActiveThemeCssFilePath();
        $cssDependencies = array(CFG_JQUERYUI_CSS, $activeTheme['cssPath'],
            CFG_PRIMEUI_CSS, CFG_FONTAWESOME_CSS, \General::getFilledMessage(CFG_ZNETDK_CSS,'primeui'),
            \General::getFilledMessage(CFG_ZNETDK_CSS,'form'));
        if (\Parameters::getPageLayoutName() !== 'custom') {
            $cssDependencies[] = \General::getFilledMessage(CFG_ZNETDK_CSS,'layout');
            $cssDependencies[] = \General::getFilledMessage(CFG_ZNETDK_CSS,'layout-'.\Parameters::getPageLayoutName());
        }
        self::addModulesDependencies('css', $cssDependencies);
        self::addApplicationDependencies('css', $cssDependencies);
        if ($activeTheme['level'] === 'application') {
            $cssDependencies[] = \General::getFilledMessage(CFG_ZNETDK_CSS,'custom-theme');
        }
        foreach ($cssDependencies as $value) {
            echo "\t".'<link rel="stylesheet" type="text/css" href="'
                    . (self::isExternalFile($value) ? '' : ZNETDK_ROOT_URI)
                    . $value . self::getFileVersion($value) . '">' . PHP_EOL;
        }
    }
        
    /**
     * Render the JavaScript dependencies in their minified version
     */
    static private function renderJS() {
        $jsDependencies = array(CFG_JQUERY_JS,CFG_JQUERYUI_JS,CFG_BLOCKUI_JS);
        $datePickerJS = \General::getFilledMessage(CFG_JQUERYUI_DATE_JS,LC_LANG_ISO_CODE);
        if (file_exists(ZNETDK_ROOT.$datePickerJS)) {
            $jsDependencies[] = $datePickerJS;
        }
        if (CFG_DEV_JS_ENABLED) {
            self::renderDevelopmentJS($jsDependencies);
        } else {
            $jsDependencies[] = CFG_PRIMEUI_JS;
            $jsDependencies[] = CFG_ZNETDK_JS;
        }
        self::addModulesDependencies('js', $jsDependencies);
        self::addApplicationDependencies('js', $jsDependencies);
        echo "\t".'<script>var znetdkAjaxURL = "'.\General::getMainScript(TRUE).'";</script>'.PHP_EOL;
        foreach ($jsDependencies as $value) {
            echo "\t".'<script src="'
                    . (self::isExternalFile($value) ? '' : ZNETDK_ROOT_URI)
                    . $value . self::getFileVersion($value) . '"></script>'.PHP_EOL;
        }
    }
    
    /**
     * Renders the JavaScript dependencies in their extended version for
     * development purpose 
     * @param array $jsDependencies Array filled with the relative path of the
     * scripts to include in the application page.
     */
    static private function renderDevelopmentJS(&$jsDependencies) {
        $extraDir = array('..', '.');
        $jsPrimeUiFiles = array_diff(scandir(ZNETDK_ROOT.CFG_PRIMEUI_JS_DEV_DIR),$extraDir);
        foreach ($jsPrimeUiFiles as $value) {
            $jsDependencies[] = CFG_PRIMEUI_JS_DEV_DIR."/".$value.'/'.$value.'.js';
        }
        $jsZnetDkFiles = array_diff(scandir(ZNETDK_ROOT.CFG_ZNETDK_JS_DIR.'/ui'),$extraDir);
        foreach ($jsZnetDkFiles as $value) {
            $jsDependencies[] = CFG_ZNETDK_JS_DIR."/ui/".$value;
        }
        $jsDependencies[] = CFG_ZNETDK_JS_DIR.'/api.js';
        $jsDependencies[] = CFG_ZNETDK_JS_DIR.'/init.js';
    }
    /**
     * Add module specific dependencies to the CSS or JS dependencies renderer.
     * @param string $subdir Subdirectory containing the files to add
     * @param array $dependencies Array containing all the dependencies to 
     * render and that is to be filled in.
     */
    static private function addModulesDependencies($subdir, &$dependencies) {
        $modules = \General::getModules();
        if ($modules === FALSE) {
            return;
        }
        $extraDir = array('..', '.','minified');
        $isMinified = FALSE;
        if ($subdir === 'css') {
            $isMinified = strstr(CFG_ZNETDK_CSS,'-min.') !== FALSE;
            $subdirectory = $isMinified ? $subdir . '/minified' : $subdir;    
        } else {
            $isMinified = CFG_DEV_JS_ENABLED === FALSE;
            $subdirectory = $isMinified ? $subdir . '/minified' : $subdir;
        }
        if ($isMinified) {
            self::minifyModuleDependencies($subdir);
        }
        foreach ($modules as $moduleName) {
            $directory = ZNETDK_MOD_ROOT . DIRECTORY_SEPARATOR . $moduleName
                    . DIRECTORY_SEPARATOR . 'public' . DIRECTORY_SEPARATOR
                    . $subdirectory;
            if (!file_exists($directory)) {
                continue;
            }
            $filesFound = array_diff(scandir($directory, SCANDIR_SORT_ASCENDING), $extraDir);
            foreach ($filesFound as $file) {
                $fileExtension = pathinfo($file, PATHINFO_EXTENSION);
                if ($fileExtension === $subdir && $file !== 'gulpfile.js') {
                    $dependencies[] = 'engine/modules/' . $moduleName . '/public/'
                        . $subdirectory . '/' . $file;
                }
            }
        }
    }
    
    static private function minifyModuleDependencies($subdir) {
        if (CFG_AUTO_MINIFICATION_METHOD === NULL) {
            return FALSE;
        }
        $method = explode('::', CFG_AUTO_MINIFICATION_METHOD);
        if (count($method) === 2 && method_exists($method[0], $method[1])) {
            $className = $method[0];
            $methodName = $method[1];
            return $className::$methodName($subdir);
        }
        return FALSE;
    }
    
    static private function addApplicationDependencies($type, &$dependencies) {
        if ($type === 'js' && CFG_APP_JS !== NULL && CFG_APP_JS !== '') {
            \ErrorHandler::suspend(); // Avoid E_NOTICE - unserialize(): Error at offset...
            $libraries = unserialize(CFG_APP_JS);
            \ErrorHandler::restart();
            if (!is_array($libraries)) {
                $libraries = array(CFG_APP_JS);
            }
            foreach ($libraries as $library) {
                $dependencies[] = $library;
            }
            return TRUE;
        } elseif ($type === 'css' && CFG_APPLICATION_CSS !== NULL && CFG_APPLICATION_CSS !== '') {
            \ErrorHandler::suspend(); // Avoid E_NOTICE - unserialize(): Error at offset...
            $libraries = unserialize(CFG_APPLICATION_CSS);
            \ErrorHandler::restart();
            if (!is_array($libraries)) {
                $libraries = array(CFG_APPLICATION_CSS);
            }
            foreach ($libraries as $library) {
                $dependencies[] = $library;
            }
            return TRUE;
        }
        return FALSE;
    }
    
    static private function isExternalFile($filePath) {
        return strpos($filePath, 'http') === 0;
    }
    
    static private function getFileVersion($filePath) {
        if (self::isExternalFile($filePath)) {
            return '';
        }
        $absoluteFilePath = ZNETDK_ROOT . $filePath;
        clearstatcache (TRUE, $absoluteFilePath);
        $fileTimestamp = filemtime($absoluteFilePath);
        return $fileTimestamp === FALSE ? '' : '?v=' . strval($fileTimestamp);
    }
    
    static private function renderCSSforMobile() {
        $cssDependencies = array(CFG_FONTAWESOME_CSS, CFG_MOBILE_W3_CSS);
        $cssDependencies[] = \General::getFilledMessage(CFG_ZNETDK_CSS, 'mobile');
        if (CFG_MOBILE_W3CSS_THEME !== NULL) {
            $cssDependencies[] = CFG_MOBILE_W3CSS_THEME;
        }
        if (CFG_MOBILE_CSS_FONT !== NULL) {
            $cssDependencies[] = CFG_MOBILE_CSS_FONT;
        }
        self::addModulesDependencies('css', $cssDependencies);
        self::addApplicationDependencies('css', $cssDependencies);
        foreach ($cssDependencies as $value) {
            echo "\t\t".'<link rel="stylesheet" type="text/css" href="'
                    . (self::isExternalFile($value) ? '' : ZNETDK_ROOT_URI)
                    . $value . self::getFileVersion($value) . '">' . PHP_EOL;
        }
    }
    
    static private function renderJSforMobile() {
        $jsDependencies = array(CFG_MOBILE_JQUERY_JS);
        if (CFG_DEV_JS_ENABLED) {
            $jsDependencies[] = CFG_ZNETDK_JS_DIR.'/mobile.js';
        } else {
            $jsDependencies[] = CFG_ZNETDK_JS_DIR.'/minified/mobile-min.js';
        }
        self::addModulesDependencies('js', $jsDependencies);
        self::addApplicationDependencies('js', $jsDependencies);
        foreach ($jsDependencies as $value) {
            echo "\t\t".'<script src="'
                    . (self::isExternalFile($value) ? '' : ZNETDK_ROOT_URI)
                    .$value . self::getFileVersion($value) . '"></script>'.PHP_EOL;
        }
    }
}